/*
	rawio.c
	Callback functions for raw I/O access to flashcart devices.
	
 Copyright (c) 2007 SukkoPera, based on work by Michael "Chishm" Chisholm.
	
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation and/or
     other materials provided with the distribution.
  3. The name of the author may not be used to endorse or promote products derived
     from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

	2007-01-08 - SukkoPera
		* Original release
*/

#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>

#include "rawio.h"
#include "file_allocation_table.h"
#include "bit_ops.h"
#include "rawio_allocate.h"

// #define DEBUG_IO
// #define WRITE_DISABLED

/* Priting from a library is silly enough, but if you know how to do it, it helps debugging :). */
#ifdef DEBUG_IO
#include <stdio.h>
#define dprintf(...) printf (__VA_ARGS__)
#else
#define dprintf(...)
#endif

/* Path will be something like "raw0:", with 0 meaning first (and only?) partition */
int _RAWIO_open_r (struct _reent *r, void *fileStruct, const char *path, int flags, int mode) {
	RAWPARTITION* partition = NULL;
	RAWIO_STRUCT* file = (RAWIO_STRUCT*) fileStruct;

	dprintf ("raw open(%s)\n", path);
	
	partition = _RAWIO_partition_getPartitionFromPath (path);
	if (partition == NULL) {
		r->_errno = ENODEV;
		return -1;
	}

	// Determine which mode the file is openned for
	if ((flags & O_CREAT) || (flags & O_EXCL) || (flags & O_TRUNC) || (flags & O_APPEND)) {
		// These flags don't make sense in raw reading
		r->_errno = EACCES;
		return -1;
	} else if ((flags & 0x03) == O_RDONLY) {
		// Open the file for read-only access
		file->read = true;
		file->write = false;
		file->append = false;
	} else if ((flags & 0x03) == O_WRONLY) {
		// Open file for write only access
		file->read = false;
		file->write = true;
		file->append = false;
	} else if ((flags & 0x03) == O_RDWR) {
		// Open file for read/write access
		file->read = true;
		file->write = true;
		file->append = false;
	} else {
		r->_errno = EACCES;
		return -1;
	}

	// Make sure we aren't trying to write to a read-only disc
	if (file->write && partition->readOnly) {
		r->_errno = EROFS;
		return -1;
	}

	file -> partition = partition;
	file -> partitionSize = partition -> numberOfSectors * partition -> bytesPerSector;
	file -> partitionStartOffset = partition -> bootSector * partition -> bytesPerSector;
	file -> currentPosition = 0;
	file -> fd = partition -> openFileCount + 10;
	partition->openFileCount += 1;

// 	dprintf ("Opened partition starting at %d, with %d sectors of %d bytes\n", partition -> fat.fatStart, partition -> numberOfSectors,
// 		 												partition -> bytesPerSector);
	
	return (int) file;
}


int _RAWIO_close_r (struct _reent *r, int fd) {
	RAWIO_STRUCT* file = (RAWIO_STRUCT*)  fd;
	
	file->partition->openFileCount -= 1;
	
	return 0;
}


int _RAWIO_read_r (struct _reent *r, int fd, char *ptr, int len) {
	RAWIO_STRUCT* file = (RAWIO_STRUCT*) fd;
	
	RAWPARTITION* partition;
	u32 realStart, startSector, endSector, sectorsToRead;
	u8 *tempBuf;
	bool done;

	dprintf ("read() %d\n", len);

	partition = file->partition;
	
	// Make sure we can actually read from the file
	if ((file == NULL) || !file->read) {
		r->_errno = EBADF;
		return -1;
	}

	// Don't try to read if the read pointer is past the end of file
	if (file->currentPosition >= file->partitionSize) {
		return 0;
	}
	
	// Don't read past end of file
	if (file->currentPosition + len > file->partitionSize) {
		r->_errno = EOVERFLOW;
		len = file->partitionSize - file->currentPosition;
	}

	// OK, do the actual read.
	realStart = file -> currentPosition + file -> partitionStartOffset;
	startSector =  realStart / partition -> bytesPerSector;
	endSector = (realStart + len - 1) / partition -> bytesPerSector;
	if (startSector >= partition -> numberOfSectors || endSector >= partition -> numberOfSectors) {
		r->_errno = EOVERFLOW;
		return 0;
	}
	sectorsToRead = endSector - startSector + 1;
	if (!(tempBuf = (u8 *) _RAWIO_mem_allocate (partition -> bytesPerSector * sectorsToRead))) {
		dprintf ("No mem\n");
		r->_errno = ENOMEM;
		return -1;
	}
	
	done = _FAT_disc_readSectors (partition->disc, startSector, sectorsToRead, tempBuf);
	if (!done) {
		dprintf ("Read err\n");
		r->_errno = EIO;
		return -1;
	} else {
		memcpy (ptr, tempBuf + (realStart % partition -> bytesPerSector), len);
	}
	_RAWIO_mem_free (tempBuf);
	
	// Update file information
	file->currentPosition += len;
	return len;
}


int _RAWIO_write_r (struct _reent *r, int fd, const char *ptr, int len) {
#ifndef WRITE_DISABLED
	RAWIO_STRUCT* file = (RAWIO_STRUCT*) fd;
	
	RAWPARTITION* partition;
	u32 realStart, startSector, endSector, sectorsToWrite;
	u8 *tempBuf;

	dprintf ("write() %d\n", len);

	partition = file->partition;

	// Make sure we can actually write to the file
	if ((file == NULL) || !file->write) {
		r->_errno = EBADF;
		return -1;
	}

	// Don't try to write if the read pointer is past the end of file
	if (file->currentPosition >= file->partitionSize) {
		r->_errno = EOVERFLOW;
		return -1;
	}
	
	// Don't write past end of file
	if (file->currentPosition + len > file->partitionSize) {
		r->_errno = EOVERFLOW;
		len = file->partitionSize - file->currentPosition;
	}

	// OK
	realStart = file -> currentPosition + file -> partitionStartOffset;
	startSector =  realStart / partition -> bytesPerSector;
	endSector = (realStart + len - 1) / partition -> bytesPerSector;
	if (startSector >= partition -> numberOfSectors || endSector >= partition -> numberOfSectors) {
		r->_errno = EOVERFLOW;
		return 0;
	}
	sectorsToWrite = endSector - startSector + 1;

	// Allocate buffer
	if (!(tempBuf = (u8 *) _RAWIO_mem_allocate (partition -> bytesPerSector * sectorsToWrite))) {
		dprintf ("No memory\n");
		r->_errno = ENOMEM;
		return -1;
	}

	// Read present data
	if (!_FAT_disc_readSectors (partition->disc, startSector, sectorsToWrite, tempBuf)) {
		dprintf ("Read error\n");
		r->_errno = EIO;
		return -1;
	}
		
	// Update buffer with data to be written */
	memcpy (tempBuf + (realStart % partition -> bytesPerSector), ptr, len);

	// Write back data
	if (!_FAT_disc_writeSectors (partition->disc, startSector, sectorsToWrite, tempBuf)) {
		dprintf ("Write error\n");
		r->_errno = EIO;
		return -1;
	}

	// Free buffer
	_RAWIO_mem_free (tempBuf);
	
	// Update file information
	file->currentPosition += len;
	
	return len;
#else
        // XXX This is a hack to make newlib printf work because it
        // doesn't like fs_pty. I'll figure out why later...
        if (fd == 1 || fd == 2) {
		return len;
	}

	r->_errno = EIO;
	return -1;
#endif
}


int _RAWIO_seek_r (struct _reent *r, int fd, int pos, int dir) {
	RAWIO_STRUCT* file = (RAWIO_STRUCT*)  fd;
	
	int position;

	if ((file == NULL)) {
		// invalid file
		r->_errno = EBADF;
		return -1;
	}

	switch (dir) {
		case SEEK_SET:
			position = pos;
			break;
		case SEEK_CUR:
			position = file->currentPosition + pos;
			break;
		case SEEK_END:	/* Makes no sense */
		default:
			r->_errno = EINVAL;
			return -1;
	}
	
	if ((pos > 0) && (position < 0)) {
		r->_errno = EOVERFLOW;
		return -1;
	}

	if (position < 0) {
		r->_errno = EINVAL;
		return -1;
	}
	
	// Only change the read/write position if it is within the bounds of the current partition
	if (position < file -> partitionSize) {
		file->currentPosition = position;
		dprintf ("fseek at %d (%d)\n", position, dir);
	} else {
		r->_errno = EINVAL;
		return -1;
	}
	
	return position;
}


int _RAWIO_fstat_r (struct _reent *r, int fd, struct stat *st) {
	dprintf ("WARNING: fstat() called\n");
	r->_errno = ENOSYS;
	return -1;
}


/* stat() and derivates don't make any sense in raw I/O. Although stat() can be used to retrieve the device type.
   We return it in the st_size member, because it has space enough. In the next devkitpro release, st_dev should be
   large enough, and a better place (see libfat). */
int _RAWIO_stat_r (struct _reent *r, const char *path, struct stat *st) {
	RAWPARTITION *partition;
	
	dprintf ("WARNING: stat() called\n");

	partition = _RAWIO_partition_getPartitionFromPath (path);
	if (partition == NULL) {
		r->_errno = ENODEV;
		return -1;
	}
	st -> st_size = _FAT_disc_hostType(partition->disc);
		
	r->_errno = ENOSYS;
	return -1;
}


int _RAWIO_link_r (struct _reent *r, const char *existing, const char *newLink) {
	dprintf ("WARNING: link() called\n");
	r->_errno = ENOTSUP;
	return -1;
}

int _RAWIO_unlink_r (struct _reent *r, const char *path) {
	dprintf ("WARNING: unlink() called\n");
	r->_errno = ENOTSUP;
	return -1;
}

int _RAWIO_chdir_r (struct _reent *r, const char *path) {
	dprintf ("WARNING: chdir() called\n");
	r->_errno = ENOSYS;
	return -1;
}
